// Canon BubbleJet BJC series driver Class
//
// 2001, Philippe Houdoin
// Based on:
// - gdevbjc.c from Aladdin GhostScript
// - LMDriver.cpp from Ryan Lockhart Lexmark 5/7xxxx BeOS driver 
//
// Printers supported
//   bj30, bj100, bj200, bj200e, bj200ex, bj230, bjc70, bjc210,
//   bjc600, bjc600e, bjc800, bjc4000 and bjc4100
// 
// Todo list:
// - color support
// - add landscape
// - decrease memory needs: multiple partial rendition?
// - gamma (global or by color) correction?

#define DEBUG	1

#include <Debug.h>

#include <stdio.h>
#include <string.h>			// for memset()

#include <StorageKit.h>

#include "PDFWriter.h"
#include "pdflib.h"


// Constructor & destructor
// ------------------------

PDFWriter::PDFWriter()
	: PrinterDriver()
{
}

PDFWriter::~PDFWriter()
{
	if ( Transport() )
		{
		CloseTransport();
		};
}


#ifdef CODEWARRIOR
	#pragma mark [Public methods]
#endif

// Public methods
// --------------

// --------------------------------------------------
status_t PDFWriter::PrintPage
	(
	int32		page_number,
	int32		page_count
	)
{
	BRect		r;
	uint32 		picture_count;
	BRect 		print_rect;
	BRect * 	pic_rects;
	BPoint * 	pic_points;
	BRegion * 	pic_region;
	BPicture ** pictures;
	uint32		i;
	
	print_rect = JobMsg()->FindRect("printable_rect");
	
	JobFile()->Read(&picture_count, sizeof(uint32));

	pictures = (BPicture **) 	malloc(picture_count * sizeof(BPicture *));
	pic_rects = (BRect *)		malloc(picture_count * sizeof(BRect));
	pic_points = (BPoint *)		malloc(picture_count * sizeof(BPoint));
	pic_region = new BRegion();
	
	for (i = 0; i < picture_count; i++)
	{
		JobFile()->Seek(40 + sizeof(off_t), SEEK_CUR);
		JobFile()->Read(&pic_points[i], sizeof(BPoint));
		JobFile()->Read(&pic_rects[i], sizeof(BRect));
		pictures[i] = new BPicture();
		pictures[i]->Unflatten(JobFile());
		pic_region->Include(pic_rects[i]);
	}
	
	r  = pic_region->Frame();
	delete pic_region;

	void *	playback_handlers[] = {
		_op0,					// 0	no operation
		_MovePenBy,				// 1	MovePenBy(void *user, BPoint delta)
		_StrokeLine,			// 2	StrokeLine(void *user, BPoint start, BPoint end)
		_StrokeRect,			// 3	StrokeRect(void *user, BRect rect)
		_FillRect,				// 4	FillRect(void *user, BRect rect)
		_StrokeRoundRect,		// 5	StrokeRoundRect(void *user, BRect rect, BPoint radii)
		_FillRoundRect,			// 6	FillRoundRect(void *user, BRect rect, BPoint radii)
		_StrokeBezier,			// 7	StrokeBezier(void *user, BPoint *control)
		_FillBezier,			// 8	FillBezier(void *user, BPoint *control)
		_StrokeArc,				// 9	StrokeArc(void *user, BPoint center, BPoint radii, float startTheta, float arcTheta)
		_FillArc,				// 10	FillArc(void *user, BPoint center, BPoint radii, float startTheta, float arcTheta)
		_StrokeEllipse,			// 11	StrokeEllipse(void *user, BPoint center, BPoint radii)
		_FillEllipse,			// 12	FillEllipse(void *user, BPoint center, BPoint radii)
		_StrokePolygon,			// 13	StrokePolygon(void *user, int32 numPoints, BPoint *points, bool isClosed)
		_FillPolygon,			// 14	FillPolygon(void *user, int32 numPoints, BPoint *points, bool isClosed)
		_op15,					// 15	*reserved*
		_op16,					// 16	*reserved*
		_DrawString,			// 17	DrawString(void *user, char *string, float deltax, float deltay)
		_DrawPixels,			// 18	DrawPixels(void *user, BRect src, BRect dest, int32 width, int32 height, int32 bytesPerRow, int32 pixelFormat, int32 flags, void *data)
		_op19,					// 19	*reserved*
		_SetClippingRects,		// 20	SetClippingRects(void *user, BRect *rects, uint32 numRects)
		_op21,					// 21	*reserved*
		_PushState,				// 22	PushState(void *user)
		_PopState,				// 23	PopState(void *user)
		_EnterStateChange,		// 24	EnterStateChange(void *user)
		_ExitStateChange,		// 25	ExitStateChange(void *user)
		_EnterFontState,		// 26	EnterFontState(void *user)
		_ExitFontState,			// 27	ExitFontState(void *user)
		_SetOrigin,				// 28	SetOrigin(void *user, BPoint pt)
		_SetPenLocation,		// 29	SetPenLocation(void *user, BPoint pt)
		_SetDrawingMode,		// 30	SetDrawingMode(void *user, drawing_mode mode)
		_SetLineMode,			// 31	SetLineMode(void *user, cap_mode capMode, join_mode joinMode, float miterLimit)
		_SetPenSize,			// 32	SetPenSize(void *user, float size)
		_SetForeColor,			// 33	SetForeColor(void *user, rgb_color color)
		_SetBackColor,			// 34	SetBackColor(void *user, rgb_color color)
		_SetStipplePattern,		// 35	SetStipplePattern(void *user, pattern p)
		_SetScale,				// 36	SetScale(void *user, float scale)
		_SetFontFamily,			// 37	SetFontFamily(void *user, char *family)
		_SetFontStyle,			// 38	SetFontStyle(void *user, char *style)
		_SetFontSpacing,		// 39	SetFontSpacing(void *user, int32 spacing)
		_SetFontSize,			// 40	SetFontSize(void *user, float size)
		_SetFontRotate,			// 41	SetFontRotate(void *user, float rotation)
		_SetFontEncoding,		// 42	SetFontEncoding(void *user, int32 encoding)
		_SetFontFlags,			// 43	SetFontFlags(void *user, int32 flags)
		_SetFontShear,			// 44	SetFontShear(void *user, float shear)
		_op45,					// 45	*reserved*
		_SetFontFace,			// 46	SetFontFace(void *user, int32 flags)
		_op47,
		_op48,
		_op49,

		NULL
	}; 

	m_log = fopen("/boot/home/Desktop/pdf_writer.log", "w");

	if ( page_number == 1 )
		{
		char	buffer[512];

		PDF_boot();
		m_pdf = PDF_new2(NULL, NULL, NULL, NULL, this);	// set *this* as pdf cookie
		PDF_open_mem(m_pdf, _WriteData);	// use callback to stream PDF document data to printer transport

		fprintf(m_log, ">>>> PDF_open_mem\n");
		
		PDF_set_parameter(m_pdf, "flush", "heavy");
		
		// find job title 
		if (JobFile()->ReadAttr("_spool/Description", B_STRING_TYPE, 0, buffer, sizeof(buffer)))
		    PDF_set_info(m_pdf, "Title", buffer);
			
		// find job creator
		if (JobFile()->ReadAttr("_spool/MimeType", B_STRING_TYPE, 0, buffer, sizeof(buffer)))
		    PDF_set_info(m_pdf, "Creator", buffer);

	    PDF_set_value(m_pdf, "compress", 0);
		};

	m_height = a4_height;
    PDF_begin_page(m_pdf, a4_width, m_height);

	fprintf(m_log, ">>>> PDF_begin_page [%f, %f]\n", a4_width, m_height);
	
	PDF_initgraphics(m_pdf);

    m_font = PDF_findfont(m_pdf, "Helvetica-Bold", "host", 0);
	m_font_size = 10;
    PDF_setfont(m_pdf, m_font, m_font_size);
 
	for (i = 0; i < picture_count; i++)
		pictures[i]->Play(playback_handlers, 50, this);

    PDF_end_page(m_pdf);

	fprintf(m_log, ">>>> PDF_end_page\n");

	if (page_number == page_count)
		{
		PDF_close(m_pdf);
		fprintf(m_log, ">>>> PDF_close\n");

    	PDF_delete(m_pdf);
	    PDF_shutdown();
		};

	fclose(m_log);

	return B_OK;
}

#ifdef CODEWARRIOR
	#pragma mark [Privates routines]
#endif

// Private routines
// ----------------

size_t PDFWriter::WriteData
	(
	void *	data,
	size_t	size
	)
{
	fprintf(m_log, ">>>> WriteData %p, %ld\n", data, size);

	return Transport()->Write(data, size);
}


void * PDFWriter::convert_bitmap
	(
	BRect 	src,
	BRect 	dest,
	int32 	width,
	int32 	height,
	int32 	bytesPerRow,
	int32 	pixelFormat,
	int32 	flags,
	void *	data
	)
{
	void *		buffer;
	uint8 * 	in;
	uint8 *		out;
	int32		x, y;
	rgb_color	color;
	
	if (pixelFormat != 8)	// B_RGB32 ?
		return NULL;

	// Reused same buffer area
	buffer 	= data;
	out 	= (uint8 *) buffer;

	for (y = (int) src.top; y <= (int) src.bottom; y++)
		{
		in = (uint8 *) data;
		in += y * bytesPerRow;
		in += 4 * (int) src.left;
	
		for (x = (int) src.left; x <= (int) src.right; x++ )
			{
			// For each pixel
			color.blue 	= *in;
			color.green	= *(in+1);
			color.red	= *(in+2);

			*out++ 	= color.red;
			*out++ = color.green;
			*out++ = color.blue;

			in += 4;	// next pixel
			};

		// next row
		};
			
	return buffer;
}


// BPicture::Play() handlers

// --------------------------------------------------
void	PDFWriter::Op
	(
	int	number
	)
{
	fprintf(m_log, "Unhandled operand %d\n", number);
}


// --------------------------------------------------
void	PDFWriter::MovePenBy
	(
	BPoint delta
	)
{
	fprintf(m_log, "MovePenBy delta=[%f, %f]\n",
			delta.x, delta.y);
}


// --------------------------------------------------
void	PDFWriter::StrokeLine
	(
	BPoint 	start,
	BPoint 	end
	)
{
	fprintf(m_log, "StrokeLine start=[%f, %f], end=[%f, %f]\n",
			start.x, start.y, end.x, end.y);
}


// --------------------------------------------------
void	PDFWriter::StrokeRect
	(
	BRect 	rect
	)
{
	fprintf(m_log, "StrokeRect rect=[%f, %f, %f, %f]\n",
			rect.left, rect.top, rect.right, rect.bottom);
}



// --------------------------------------------------
void	PDFWriter::FillRect
	(
	BRect 	rect
	)
{
	fprintf(m_log, "FillRect rect=[%f, %f, %f, %f]\n",
			rect.left, rect.top, rect.right, rect.bottom);
}


// --------------------------------------------------
void	PDFWriter::StrokeRoundRect
	(
	BRect 	rect,
	BPoint	radii
	)
{
	fprintf(m_log, "StrokeRoundRect center=[%f, %f, %f, %f], radii=[%f, %f]\n",
			rect.left, rect.top, rect.right, rect.bottom, radii.x, radii.y);
}


// --------------------------------------------------
void	PDFWriter::FillRoundRect
	(
	BRect 	rect,
	BPoint	radii
	)
{
	fprintf(m_log, "FillRoundRect rect=[%f, %f, %f, %f], radii=[%f, %f]\n",
			rect.left, rect.top, rect.right, rect.bottom, radii.x, radii.y);
}


// --------------------------------------------------
void	PDFWriter::StrokeBezier
	(
	BPoint	* control
	)
{
	fprintf(m_log, "StrokeBezier\n");
}


// --------------------------------------------------
void	PDFWriter::FillBezier
	(
	BPoint	* control
	)
{
	fprintf(m_log, "FillBezier\n");
}


// --------------------------------------------------
void	PDFWriter::StrokeArc
	(
	BPoint 	center,
	BPoint 	radii,
	float 	startTheta,
	float 	arcTheta
	)
{
	fprintf(m_log, "StrokeArc center=[%f, %f], radii=[%f, %f], startTheta=%f, arcTheta=%f\n",
			center.x, center.y, radii.x, radii.y, startTheta, arcTheta);
}


// --------------------------------------------------
void	PDFWriter::FillArc
	(
	BPoint 	center,
	BPoint 	radii,
	float 	startTheta,
	float 	arcTheta
	)
{
	fprintf(m_log, "FillArc center=[%f, %f], radii=[%f, %f], startTheta=%f, arcTheta=%f\n",
			center.x, center.y, radii.x, radii.y, startTheta, arcTheta);
}


// --------------------------------------------------
void	PDFWriter::StrokeEllipse
	(
	BPoint 	center,
	BPoint 	radii
	)
{
	fprintf(m_log, "StrokeEllipse center=[%f, %f], radii=[%f, %f]\n",
			center.x, center.y, radii.x, radii.y);
}


// --------------------------------------------------
void	PDFWriter::FillEllipse
	(
	BPoint 	center,
	BPoint 	radii
	)
{
	fprintf(m_log, "FillEllipse center=[%f, %f], radii=[%f, %f]\n",
			center.x, center.y, radii.x, radii.y);
}


// --------------------------------------------------
void	PDFWriter::StrokePolygon
	(
	int32 		numPoints,
	BPoint *	points,
	bool 		isClosed
	)
{
	int32	i;
	
	fprintf(m_log, "StrokePolygon numPoints=%ld, isClosed=%d\npoints=",
			numPoints, isClosed);
			
	for ( i = 0; i < numPoints; i++, points++ )
		fprintf(m_log, " [%f, %f]", points->x, points->y);
	fprintf(m_log, "\n");
}


// --------------------------------------------------
void	PDFWriter::FillPolygon
	(
	int32 		numPoints,
	BPoint *	points,
	bool 		isClosed
	)
{
	int32	i;
	
	fprintf(m_log, "FillPolygon numPoints=%ld, isClosed=%dpoints=\n",
			numPoints, isClosed);
			
	for ( i = 0; i < numPoints; i++, points++ )
		fprintf(m_log, " [%f, %f]", points->x, points->y);
	fprintf(m_log, "\n");
}


// --------------------------------------------------
void	PDFWriter::DrawString
	(
	char *	string,
	float 	deltax,
	float 	deltay
	)
{
	fprintf(m_log, "DrawString string=\"%s\", deltax=%f, deltay=%f, at %f, %f\n",
			string, deltax, deltay, m_pen_x, m_pen_y);
			
    PDF_setfont(m_pdf, m_font, m_font_size);
    PDF_set_text_pos(m_pdf, m_pen_x, m_pen_y);
	PDF_show(m_pdf, string);
}


// --------------------------------------------------
void	PDFWriter::DrawPixels
	(
	BRect 	src,
	BRect 	dest,
	int32 	width,
	int32 	height,
	int32 	bytesPerRow,
	int32 	pixelFormat,
	int32 	flags,
	void *	data
	)
{
	void * 	buffer;
	int		w, h;
	int 	image;

	fprintf(m_log, "DrawPixels src=[%f, %f, %f, %f], dest=[%f, %f, %f, %f], "
					"width=%ld, height=%ld, bytesPerRow=%ld, pixelFormat=%ld, "
					"flags=%ld, data=%p\n",
					src.left, src.top, src.right, src.bottom,
					dest.left, dest.top, dest.right, dest.bottom,
					width, height, bytesPerRow, pixelFormat, flags, data);

	buffer = convert_bitmap(src, dest, width, height, bytesPerRow, pixelFormat, flags, data);
	
	if (! buffer)
		return;
		
	w = (int) (src.right - src.left + 1);
	h = (int) (src.bottom - src.top + 1);
		
	image = PDF_open_image(m_pdf, "raw", "memory", (const char *) buffer, w*h*3, w, h, 3, 8, 0);
	if ( image >= 0 )
		{
		float 	x, y;

		x = m_x0 + dest.left;
		y = m_height - (m_y0 + dest.bottom);
		
		PDF_place_image(m_pdf, image, x, y, 1.0);
		PDF_close_image(m_pdf, image);
		};
	
	if (buffer != data)
		free(buffer);
}


// --------------------------------------------------
void	PDFWriter::SetClippingRects
	(
	BRect *	rects,
	uint32 	numRects
	)
{
	uint32	i;
	
	fprintf(m_log, "SetClippingRects numRects=%ld\nrects=",
			numRects);
			
	for ( i = 0; i < numRects; i++, rects++ )
		fprintf(m_log, " [%f, %f, %f, %f]",
				rects->left, rects->top, rects->right, rects->bottom);
	fprintf(m_log, "\n");
}


// --------------------------------------------------
void	PDFWriter::PushState()
{
	fprintf(m_log, "PushState\n");

	PDF_save(m_pdf);
}


// --------------------------------------------------
void	PDFWriter::PopState()
{
	fprintf(m_log, "PopState\n");
	
	PDF_restore(m_pdf);
}


// --------------------------------------------------
void	PDFWriter::EnterStateChange()
{
	fprintf(m_log, "EnterStateChange\n");
}


// --------------------------------------------------
void	PDFWriter::ExitStateChange()
{
	fprintf(m_log, "ExitStateChange\n");
}


// --------------------------------------------------
void	PDFWriter::EnterFontState()
{
	fprintf(m_log, "EnterFontState\n");
}


// --------------------------------------------------
void	PDFWriter::ExitFontState()
{
	fprintf(m_log, "ExitFontState\n");
}


// --------------------------------------------------
void	PDFWriter::SetOrigin
	(
	BPoint 	pt
	)
{
	fprintf(m_log, "SetOrigin pt=[%f, %f]\n",
			pt.x, pt.y);
			
	m_x0 = pt.x;
	m_y0 = pt.y;
}


// --------------------------------------------------
void	PDFWriter::SetPenLocation
	(
	BPoint 	pt
	)
{
	fprintf(m_log, "SetPenLocation pt=[%f, %f]\n",
			pt.x, pt.y);
		
	m_pen_x = m_x0 + pt.x;
	m_pen_y	= m_height - (m_y0 + pt.y);
	
	// PDF_moveto(m_pdf, m_x0 + pt.x, m_height - (m_y0 + pt.y));
	// PDF_set_text_pos(m_pdf, m_x0 + pt.x, m_height - (m_y0 + pt.y));
}



// --------------------------------------------------
void	PDFWriter::SetDrawingMode
	(
	drawing_mode 	mode
	)
{
	fprintf(m_log, "SetDrawingMode mode=%d\n", mode);
}



// --------------------------------------------------
void	PDFWriter::SetLineMode
	(
	cap_mode 	capMode,
	join_mode 	joinMode,
	float 		miterLimit
	)
{
	fprintf(m_log, "SetLineMode\n");
}


// --------------------------------------------------
void	PDFWriter::SetPenSize
	(
	float size
	)
{
	fprintf(m_log, "SetPenSize size=%f\n", size);

	PDF_setlinewidth(m_pdf, size);
}


// --------------------------------------------------
void	PDFWriter::SetForeColor
	(
	rgb_color 	color
	)
{
	fprintf(m_log, "SetForColor color=[%d, %d, %d, %d]\n",
			color.red, color.green, color.blue, color.alpha);

	PDF_setcolor(m_pdf, "both", "rgb", color.red / 255, color.green / 255, color.blue / 255, 0.0);
}


// --------------------------------------------------
void	PDFWriter::SetBackColor
	(
	rgb_color 	color
	)
{
	fprintf(m_log, "SetBackColor color=[%d, %d, %d, %d]\n",
			color.red, color.green, color.blue, color.alpha);

	// PDF_setcolor(m_pdf, "fill", "rgb", color.red / 255, color.green / 255, color.blue / 255, 0.0);
}


// --------------------------------------------------
void	PDFWriter::SetStipplePattern
	(
	pattern 	pat
	)
{
	fprintf(m_log, "SetStipplePattern\n");
}


// --------------------------------------------------
void	PDFWriter::SetScale
	(
	float scale
	)
{
	fprintf(m_log, "SetScale scale=%f\n", scale);

}


// --------------------------------------------------
void	PDFWriter::SetFontFamily
	(
	char *	family
	)
{
	fprintf(m_log, "SetFontFamily family=\"%s\"\n", family);
}


// --------------------------------------------------
void	PDFWriter::SetFontStyle
	(
	char *	style
	)
{
	fprintf(m_log, "SetFontStyle style=\"%s\"\n", style);
}


// --------------------------------------------------
void	PDFWriter::SetFontSpacing
	(
	int32 	spacing
	)
{
	fprintf(m_log, "SetFontSpacing spacing=%ld\n", spacing);
}


// --------------------------------------------------
void	PDFWriter::SetFontSize
	(
	float size
	)
{
	fprintf(m_log, "SetFontSize size=%f\n", size);
	
	m_font_size = size;
}


// --------------------------------------------------
void	PDFWriter::SetFontRotate
	(
	float 	rotation
	)
{
	fprintf(m_log, "SetFontRotate rotation=%f\n", rotation);
}


// --------------------------------------------------
void	PDFWriter::SetFontEncoding
	(
	int32 	encoding
	)
{
	fprintf(m_log, "SetFontEncoding encoding=%ld\n", encoding);
}


// --------------------------------------------------
void	PDFWriter::SetFontFlags
	(
	int32	flags
	)
{
	fprintf(m_log, "SetFontFlags flags=%ld (0x%lx)\n", flags, flags);
}


// --------------------------------------------------
void	PDFWriter::SetFontShear
	(
	float 	shear
	)
{
	fprintf(m_log, "SetFontShear shear=%f\n", shear);
}


// --------------------------------------------------
void	PDFWriter::SetFontFace
	(
	int32 	flags
	)
{
	fprintf(m_log, "SetFontFace flags=%ld\n", flags);
}


#ifdef CODEWARRIOR
	#pragma mark [Redirectors to instance callbacks/handlers]
#endif

// Private routines
// ----------------


size_t	_WriteData(PDF * pdf, void * data, size_t size)	{ return ((PDFWriter *) PDF_get_opaque(pdf))->WriteData(data, size); }

// BPicture playback handlers class instance redirectors
void		_MovePenBy(void * p, BPoint delta) 														{ return ((PDFWriter *) p)->MovePenBy(delta); }
void		_StrokeLine(void * p, BPoint start, BPoint end) 										{ return ((PDFWriter *) p)->StrokeLine(start, end); }
void		_StrokeRect(void * p, BRect rect) 														{ return ((PDFWriter *) p)->StrokeRect(rect); }
void		_FillRect(void * p, BRect rect) 														{ return ((PDFWriter *) p)->FillRect(rect); }
void		_StrokeRoundRect(void * p, BRect rect, BPoint radii) 									{ return ((PDFWriter *) p)->StrokeRoundRect(rect, radii); }
void		_FillRoundRect(void * p, BRect rect, BPoint radii)  									{ return ((PDFWriter *) p)->FillRoundRect(rect, radii); }
void		_StrokeBezier(void * p, BPoint *control)  												{ return ((PDFWriter *) p)->StrokeBezier(control); }
void		_FillBezier(void * p, BPoint *control)  												{ return ((PDFWriter *) p)->FillBezier(control); }
void		_StrokeArc(void * p, BPoint center, BPoint radii, float startTheta, float arcTheta)		{ return ((PDFWriter *) p)->StrokeArc(center, radii, startTheta, arcTheta); }
void		_FillArc(void * p, BPoint center, BPoint radii, float startTheta, float arcTheta)		{ return ((PDFWriter *) p)->FillArc(center, radii, startTheta, arcTheta); }
void		_StrokeEllipse(void * p, BPoint center, BPoint radii)									{ return ((PDFWriter *) p)->StrokeEllipse(center, radii); }
void		_FillEllipse(void * p, BPoint center, BPoint radii)										{ return ((PDFWriter *) p)->FillEllipse(center, radii); }
void		_StrokePolygon(void * p, int32 numPoints, BPoint *points, bool isClosed) 				{ return ((PDFWriter *) p)->StrokePolygon(numPoints, points, isClosed); }
void		_FillPolygon(void * p, int32 numPoints, BPoint *points, bool isClosed)					{ return ((PDFWriter *) p)->FillPolygon(numPoints, points, isClosed); }
void		_DrawString(void * p, char *string, float deltax, float deltay)							{ return ((PDFWriter *) p)->DrawString(string, deltax, deltay); }
void		_DrawPixels(void * p, BRect src, BRect dest, int32 width, int32 height, int32 bytesPerRow, int32 pixelFormat, int32 flags, void *data)
						{ return ((PDFWriter *) p)->DrawPixels(src, dest, width, height, bytesPerRow, pixelFormat, flags, data); }
void		_SetClippingRects(void * p, BRect *rects, uint32 numRects)								{ return ((PDFWriter *) p)->SetClippingRects(rects, numRects); }
void		_PushState(void * p)  																	{ return ((PDFWriter *) p)->PushState(); }
void		_PopState(void * p)  																	{ return ((PDFWriter *) p)->PopState(); }
void		_EnterStateChange(void * p) 															{ return ((PDFWriter *) p)->EnterStateChange(); }
void		_ExitStateChange(void * p) 																{ return ((PDFWriter *) p)->ExitStateChange(); }
void		_EnterFontState(void * p) 																{ return ((PDFWriter *) p)->EnterFontState(); }
void		_ExitFontState(void * p) 																{ return ((PDFWriter *) p)->ExitFontState(); }
void		_SetOrigin(void * p, BPoint pt)															{ return ((PDFWriter *) p)->SetOrigin(pt); }
void		_SetPenLocation(void * p, BPoint pt)													{ return ((PDFWriter *) p)->SetPenLocation(pt); }
void		_SetDrawingMode(void * p, drawing_mode mode)											{ return ((PDFWriter *) p)->SetDrawingMode(mode); }
void		_SetLineMode(void * p, cap_mode capMode, join_mode joinMode, float miterLimit)			{ return ((PDFWriter *) p)->SetLineMode(capMode, joinMode, miterLimit); }
void		_SetPenSize(void * p, float size)														{ return ((PDFWriter *) p)->SetPenSize(size); }
void		_SetForeColor(void * p, rgb_color color)												{ return ((PDFWriter *) p)->SetForeColor(color); }
void		_SetBackColor(void * p, rgb_color color)												{ return ((PDFWriter *) p)->SetBackColor(color); }
void		_SetStipplePattern(void * p, pattern pat)												{ return ((PDFWriter *) p)->SetStipplePattern(pat); }
void		_SetScale(void * p, float scale)														{ return ((PDFWriter *) p)->SetScale(scale); }
void		_SetFontFamily(void * p, char *family)													{ return ((PDFWriter *) p)->SetFontFamily(family); }
void		_SetFontStyle(void * p, char *style)													{ return ((PDFWriter *) p)->SetFontStyle(style); }
void		_SetFontSpacing(void * p, int32 spacing)												{ return ((PDFWriter *) p)->SetFontSpacing(spacing); }
void		_SetFontSize(void * p, float size)														{ return ((PDFWriter *) p)->SetFontSize(size); }
void		_SetFontRotate(void * p, float rotation)												{ return ((PDFWriter *) p)->SetFontRotate(rotation); }
void		_SetFontEncoding(void * p, int32 encoding)												{ return ((PDFWriter *) p)->SetFontEncoding(encoding); }
void		_SetFontFlags(void * p, int32 flags)													{ return ((PDFWriter *) p)->SetFontFlags(flags); }
void		_SetFontShear(void * p, float shear)													{ return ((PDFWriter *) p)->SetFontShear(shear); }
void		_SetFontFace(void * p, int32 flags)														{ return ((PDFWriter *) p)->SetFontFace(flags); }

// undefined or undocumented operation handlers...
void		_op0(void * p)	{ return ((PDFWriter *) p)->Op(0); }
void		_op15(void * p)	{ return ((PDFWriter *) p)->Op(15); }
void		_op16(void * p)	{ return ((PDFWriter *) p)->Op(16); }
void		_op19(void * p)	{ return ((PDFWriter *) p)->Op(19); }
void		_op21(void * p)	{ return ((PDFWriter *) p)->Op(21); }
void		_op45(void * p)	{ return ((PDFWriter *) p)->Op(45); }
void		_op47(void * p)	{ return ((PDFWriter *) p)->Op(47); }
void		_op48(void * p)	{ return ((PDFWriter *) p)->Op(48); }
void		_op49(void * p)	{ return ((PDFWriter *) p)->Op(49); }


