#ifndef PDFWRITER_H
#define PDFWRITER_H

#include <AppKit.h>
#include <InterfaceKit.h>

#include "PrinterDriver.h"

#include "pdflib.h"

#define SCALE 8.3333f
#define SCREEN 72.0f
#define LETTER_WIDTH 8.5
#define LETTER_HEIGHT 11

class PDFWriter : public PrinterDriver
	{
	public:
		// constructors / destructor
		PDFWriter();
		~PDFWriter();
		
		// public methods
		status_t 			PrintPage(int32 page_number, int32 page_count);
		
		// PDFLib callbacks
		size_t		WriteData(void * data, size_t size);
		
		void * 		convert_bitmap(BRect src, BRect dest, int32 width, int32 height, int32 bytesPerRow, int32 pixelFormat, int32 flags, void * in);

		// BPicture playback handlers
		void		Op(int number);
		void		MovePenBy(BPoint delta);
		void		StrokeLine(BPoint start, BPoint end);
		void		StrokeRect(BRect rect);
		void		FillRect(BRect rect);
		void		StrokeRoundRect(BRect rect, BPoint radii);
		void		FillRoundRect(BRect rect, BPoint radii);
		void		StrokeBezier(BPoint *control);
		void		FillBezier(BPoint *control);
		void		StrokeArc(BPoint center, BPoint radii, float startTheta, float arcTheta);
		void		FillArc(BPoint center, BPoint radii, float startTheta, float arcTheta);
		void		StrokeEllipse(BPoint center, BPoint radii);
		void		FillEllipse(BPoint center, BPoint radii);
		void		StrokePolygon(int32 numPoints, BPoint *points, bool isClosed);
		void		FillPolygon(int32 numPoints, BPoint *points, bool isClosed);
		void		DrawString(char *string, float deltax, float deltay);
		void		DrawPixels(BRect src, BRect dest, int32 width, int32 height, int32 bytesPerRow, int32 pixelFormat, int32 flags, void *data);
		void		SetClippingRects(BRect *rects, uint32 numRects);
		void		PushState();
		void		PopState();
		void		EnterStateChange();
		void		ExitStateChange();
		void		EnterFontState();
		void		ExitFontState();
		void		SetOrigin(BPoint pt);
		void		SetPenLocation(BPoint pt);
		void		SetDrawingMode(drawing_mode mode);
		void		SetLineMode(cap_mode capMode, join_mode joinMode, float miterLimit);
		void		SetPenSize(float size);
		void		SetForeColor(rgb_color color);
		void		SetBackColor(rgb_color color);
		void		SetStipplePattern(pattern p);
		void		SetScale(float scale);
		void		SetFontFamily(char *family);
		void		SetFontStyle(char *style);
		void		SetFontSpacing(int32 spacing);
		void		SetFontSize(float size);
		void		SetFontRotate(float rotation);
		void		SetFontEncoding(int32 encoding);
		void		SetFontFlags(int32 flags);
		void		SetFontShear(float shear);
		void		SetFontFace(int32 flags);
		
	private:
		FILE *			m_log;
		PDF *			m_pdf;
		int				m_font;
		float			m_font_size;
		float			m_height;
		float			m_x0;
		float			m_y0;
		float			m_pen_x;
		float			m_pen_y;
	};


// PDFLib C callbacks class instance redirectors
size_t		_WriteData(PDF * p, void * data, size_t size);

// BPicture playback handlers class instance redirectors
void		_MovePenBy(void * p, BPoint delta);
void		_StrokeLine(void * p, BPoint start, BPoint end);
void		_StrokeRect(void * p, BRect rect);
void		_FillRect(void * p, BRect rect);
void		_StrokeRoundRect(void * p, BRect rect, BPoint radii);
void		_FillRoundRect(void * p, BRect rect, BPoint radii);
void		_StrokeBezier(void * p, BPoint * control);
void		_FillBezier(void * p, BPoint * control);
void		_StrokeArc(void * p, BPoint center, BPoint radii, float startTheta, float arcTheta);
void		_FillArc(void * p, BPoint center, BPoint radii, float startTheta, float arcTheta);
void		_StrokeEllipse(void * p, BPoint center, BPoint radii);
void		_FillEllipse(void * p, BPoint center, BPoint radii);
void		_StrokePolygon(void * p, int32 numPoints, BPoint * points, bool isClosed);
void		_FillPolygon(void * p, int32 numPoints, BPoint * points, bool isClosed);
void		_DrawString(void * p, char * string, float deltax, float deltay);
void		_DrawPixels(void * p, BRect src, BRect dest, int32 width, int32 height, int32 bytesPerRow, int32 pixelFormat, int32 flags, void * data);
void		_SetClippingRects(void * p, BRect *rects, uint32 numRects);
void		_PushState(void * p);
void		_PopState(void * p);
void		_EnterStateChange(void * p);
void		_ExitStateChange(void * p);
void		_EnterFontState(void * p);
void		_ExitFontState(void * p);
void		_SetOrigin(void * p, BPoint pt);
void		_SetPenLocation(void * p, BPoint pt);
void		_SetDrawingMode(void * p, drawing_mode mode);
void		_SetLineMode(void * p, cap_mode capMode, join_mode joinMode, float miterLimit);
void		_SetPenSize(void * p, float size);
void		_SetForeColor(void * p, rgb_color color);
void		_SetBackColor(void * p, rgb_color color);
void		_SetStipplePattern(void * p, pattern pat);
void		_SetScale(void * p, float scale);
void		_SetFontFamily(void * p, char * family);
void		_SetFontStyle(void * p, char * style);
void		_SetFontSpacing(void * p, int32 spacing);
void		_SetFontSize(void * p, float size);
void		_SetFontRotate(void * p, float rotation);
void		_SetFontEncoding(void * p, int32 encoding);
void		_SetFontFlags(void * p, int32 flags);
void		_SetFontShear(void * p, float shear);
void		_SetFontFace(void * p, int32 flags);

// undefined or undocumented operation handlers...
void		_op0(void * p);
void		_op15(void * p);
void		_op16(void * p);
void		_op19(void * p);
void		_op21(void * p);
void		_op45(void * p);
void		_op47(void * p);
void		_op48(void * p);
void		_op49(void * p);

#endif	// #if PDFWRITER_H
