#ifndef PRINTERDRIVER_H
#define PRINTERDRIVER_H

#include <AppKit.h>
#include <InterfaceKit.h>

#ifndef ROUND_UP
	#define ROUND_UP(x, y) (((x) + (y) - 1) & ~((y) - 1))
#endif

// transport add-on calls definition 
extern "C"
	{
	typedef BDataIO * 	(*init_transport_proc)(BMessage *);
	typedef void 		(*exit_transport_proc)(void);
	};	


class PrinterDriver
	{
	public:
		// constructors / destructor
								PrinterDriver();
		virtual					~PrinterDriver();
		
		virtual status_t 		PrintJob(BFile * job_file, BNode * printer_node, BMessage * job_msg);
		virtual status_t		PrintPage(int32 page_number, int32 page_count);

		// configuration default methods
		virtual status_t 		PrinterSetup(char * printer_name);
		virtual status_t 		PageSetup(BMessage * msg);
		virtual status_t 		JobSetup(BMessage * msg);
		
		// transport-related methods
		status_t				OpenTransport();
		status_t				CloseTransport();

		// accessors
		inline BFile *			JobFile()		{ return m_job_file; }
		inline BNode *			PrinterNode()	{ return m_printer_node; }
		inline BMessage *		JobMsg()		{ return m_job_msg; }
		inline BDataIO *		Transport()		{ return m_transport; }
		
		// publics status code
		typedef enum {
			PORTRAIT_ORIENTATION,
			LANDSCAPE_ORIENTATION
		} Orientation;

	private:
		BFile *					m_job_file;
		BNode *					m_printer_node;
		BMessage *				m_job_msg;

		volatile Orientation	m_orientation;
		
		// transport-related 
		BDataIO *				m_transport;
		image_id				m_transport_add_on;
		init_transport_proc		m_transport_init_proc;
		exit_transport_proc		m_transport_exit_proc;
	};

#endif // #ifndef PRINTERDRIVER_H

